import _ from 'lodash'
import chalk from 'chalk'
import c from 'chalk'
import cl from 'chalkline'
import path, {join} from 'path'
import {getColorForDuration, ms, printHelp} from './util'
import indent from 'indent-string'

export default function printDupes({requestToFilename, config}) {
  printHelp('Duplicates', getHelpStr())

  _(requestToFilename)
    .map((v, req) => {
      if (Object.keys(v).length <= 1) return
      const requestToFilenames = _(v)
        .map((v, k) => {
          const filename = k
          const {cached, duration, childrenDuration, childrenResolveTime} = v.node.model
          const parentName = v.node.parent.model.filename
          return {filename, duration, cached, parentName, childrenDuration, childrenResolveTime}
        })
        .value()

      // Here we can add up all the durations of the cached misses
      //   to reveal how much time can be saved.
      const wastedTime = _(requestToFilenames)
        //.omitBy('cached')
        //.tail() // We must require it once, so assume the first non-cached call is how long it would take.
        .reduce((acc, v) => {
          return acc + v.duration
        }, 0)

      //////////////////////////////////////////////////////////////////////////
      // Render

      const timeStr = ms(wastedTime, getColorForDuration(wastedTime))
      console.log(req + ' ' + timeStr)
      _(requestToFilenames).each(({filename, duration, cached, parentName, childrenDuration}) => {
        const relFilename = path.relative(process.cwd(), filename)
        const relParentFilename = path.relative(process.cwd(), parentName)
        const color = cached ? 'green' : 'white'
        let filenameStr =
          _.get(chalk, color)(filename) +
          ' '
        const totalDuration = duration + childrenDuration
        filenameStr += ms(totalDuration, getColorForDuration(totalDuration))
        let parentStr = c.grey.dim(relParentFilename)
        const cachedStr = cached ? 'hit' : 'miss'
        console.log(indent(filenameStr + ' ' + cachedStr, 2))
        if (config.printDupesShowRequester)
          console.log(indent(parentStr, 4))
      })

      //////////////////////////////////////////////////////////////////////////
    })
    .value()
}

////////////////////////////////////////////////////////////////////////////////

function getHelpStr() {
  return `
  Show which files made the same request, and whether this request used a cached module from 
  'require.cache', or instantiated a fresh module. 
  
  A cache miss is caused by the same request resolving to a different absolute file path.
  
  When using the deduping functionality, a miss is caused by:
  
  - the version requested in the package.json file was not satisfied by a root dep, or a pnpm store dep.
    - FIX: update the deps package.json to a later version.
  
  When not using the deduping functionality, a miss is caused by:
  
  - an installed module has the dependency in its node_modules folder and will use that copy. (This should never happen with npm3)
    - FIX: 
      - npm2: npm dedupe
      - npm3: problem should not occur
      - enable live-perf deduper
  
  Entries are excluded if there was only one request because it would always be a cache miss.
  
  Look for dupes that take a long time to require - this can lead to big savings.
  `
}

////////////////////////////////////////////////////////////////////////////////

// `E.g.
//
//     ˅ total time spent requiring
// ms 2ms
// ˄ requested package i.e. require('ms') / import 'ms'                      ˅ time spent requiring
//   /xxx/node_modules/.registry.npmjs.org/ms/2.1.1/node_modules/ms/index.js 0ms hit
//   ˄ resolve path                                                              ˄ cache hit/miss
//     node_modules/.registry.npmjs.org/debug/4.1.0/node_modules/debug/src/node.js
//     ˄ requester - file containing require/import statement
//   /xxx/node_modules/.registry.npmjs.org/ms/2.0.0/node_modules/ms/index.js 1ms miss
//     node_modules/.registry.npmjs.org/debug/2.6.9/node_modules/debug/src/debug.js
// `
